"""Dataset schema classes."""

from collections.abc import Callable, Iterable
from datetime import datetime
from enum import Enum
from typing import Any, Generic, Self, TypeVar, get_args, get_origin

from pandera import Column, DataFrameModel
from pandera.dtypes import DataType, is_numeric, is_string
from pandera.typing.common import DataFrameBase
from pydantic import BaseModel, Field, create_model

DataFrameBaseT = TypeVar("DataFrameBaseT", bound=DataFrameBase)
DatasetFunctionType = Callable[[], DataFrameBaseT]


# Transforming DataFrameModel to BaseModel


class ValueTypeEnum(str, Enum):
    """The type of a value in a column. Used for fabricating Pydantic models."""

    STRING = "string"
    NUMBER = "number"

    @classmethod
    def from_pandera_dtype(cls, dtype: DataType) -> "ValueTypeEnum":
        """Convert a pandera dtype to a ValueTypeEnum."""
        if is_numeric(dtype):
            return cls.NUMBER
        if is_string(dtype):
            return cls.STRING
        msg = f"Unsupported dtype: {dtype}"
        raise ValueError(msg)

    def into_python_type(self) -> type[str] | type[float]:
        """Convert a ValueTypeEnum to a Python type."""
        match self:
            case self.STRING:
                return str
            case self.NUMBER:
                return float
            case _:
                msg = f"Unsupported value type: {self}"
                raise ValueError(msg)


def create_column_base_model_from_dataframe_model(
    dataframe_model: type[DataFrameModel],
) -> type[BaseModel]:
    """Create a BaseModel from a DataFrameModel.

    ### Arguments:
    - dataframe_model: The DataFrameModel to create a BaseModel from.

    ### Returns:
    The BaseModel class.
    """
    schema_columns: Iterable[Column] = dataframe_model.to_schema().columns.values()
    fields: dict[str, tuple[type[Any], Any]] = {}
    for col in schema_columns:
        python_type = ValueTypeEnum.from_pandera_dtype(col.dtype).into_python_type()
        fields[col.name] = (python_type, Field(description=col.description))

    return create_model(f"{dataframe_model.__name__}Model", **fields)


# Response returned by all view endpoints

M = TypeVar("M", bound=BaseModel | list[BaseModel])


class ViewEndpointResponseModel(BaseModel, Generic[M]):
    """The return object of a view endpoint."""

    last_updated: datetime | None
    """The last updated date of the view."""

    data: M | None
    """The data of the view."""

    @classmethod
    def create_class_from_model(
        cls: type[Self], model: type[M]
    ) -> type["ViewEndpointResponseModel[M]"]:
        """Create a response wrapper class from a model.

        ### Arguments:
        - model: The model to create a response wrapper class from.

        ### Returns:
        The response wrapper class.
        """
        # If the model is NOT a list, we can just use its properties directly
        base_model = model
        name = base_model.__name__

        # If it is a list, we need to extract the base model from inside the list
        is_list = get_origin(model) is list
        if is_list:
            args = get_args(model)
            if len(args) != 1:
                msg = f"List model must have exactly one type argument, got {len(args)}"
                raise ValueError(msg)
            base_model = args[0]
            name = f"{base_model.__name__}List"

        fields = {
            "last_updated": (
                datetime | None,
                Field(description="The time the view was last updated at."),
            ),
            "data": (model | None, Field(description="The data of the view.")),
        }
        return create_model(
            f"{name}Response",
            **fields,
            __doc__=f"Response container model for {name} (data).",
        )



"""Base view class."""

from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from functools import cached_property
from typing import Generic
from uuid import UUID, uuid4

from pandera.polars import DataFrameModel
from pandera.typing.polars import DataFrame
from pydantic import BaseModel

from tacobi.data_model.models import DataModelType
from tacobi.view.type_utils import (
    extract_container_type,
    extract_model_from_typehints,
    extract_return_type,
)
from tacobi.view.view_models.endpoint_model import (
    ViewEndpointResponseModel,
    create_column_base_model_from_dataframe_model,
)


@dataclass
class BaseView(Generic[DataModelType]):
    """Base view class."""

    function: Callable[..., Awaitable[DataModelType]]
    """The function to call to update the view."""

    dependencies: list[UUID] = field(default_factory=list)
    """The dependencies of the view used for topological sorting."""

    route: str | None = None
    """The optional REST route of the view."""

    name: str | None = None
    """The optional name of the view for logging and debugging."""

    id: UUID = field(default_factory=uuid4)
    """The unique identifier for the view."""

    def __str__(self) -> str:
        """Get the string representation of the view."""
        return f"View(name={self.name}, id={self.id})"

    def __hash__(self) -> int:
        """Hash the view."""
        return hash(self.id)

    # =====================================================
    # Pydantic BaseModel conversion
    # =====================================================

    @cached_property
    def return_type(self) -> type[DataModelType]:
        """The return type of the view."""
        return extract_return_type(self.function)

    @cached_property
    def base_model(self) -> tuple[type[BaseModel], bool]:
        """The base model of the view and whether it's a list (aka a DataFrameModel).

        ### Returns:
        A tuple of the base model and a boolean indicating whether it's a list.
        """
        # If it's already a BaseModel, we can return it directly
        if issubclass(self.return_type, BaseModel):
            return self.return_type, False

        # Otherwise, it's a DataFrameModel, so we need to wrap it in a list
        container_type = extract_container_type(self.return_type, DataFrame)
        model = extract_model_from_typehints(
            self.function, container_type, DataFrameModel
        )

        return create_column_base_model_from_dataframe_model(model), True

    @cached_property
    def fastapi_response_model(self) -> type[ViewEndpointResponseModel]:
        """The response model for the view.

        ### Returns:
        The response model for the view.
        """
        base_model, is_list = self.base_model
        model = base_model if not is_list else list[base_model]
        return ViewEndpointResponseModel.create_class_from_model(model)

    def convert_to_base_model(self, data: DataModelType) -> BaseModel | list[BaseModel]:
        """Convert data to BaseModel format.

        Called internally when generating the FastAPI response.

        ### Arguments:
        - data: Either a BaseModel instance or DataFrame to convert

        ### Returns:
        BaseModel instance or list of BaseModel instances
        """
        # If it's already a BaseModel, return it directly
        if isinstance(data, BaseModel):
            return data

        # Otherwise, assume it's a DataFrame and convert to list of BaseModels
        base_model_class, _ = self.base_model

        # Handle both pandas and polars DataFrames
        if hasattr(data, "to_dicts"):  # Polars
            records = data.to_dicts()
        else:  # Pandas
            records = data.to_dict("records")

        return [base_model_class(**row) for row in records]



"""Materialized and normal views that can be used to query cached data."""

from tacobi.view.view_models.base import BaseView
from tacobi.view.view_models.materialized import MaterializedView
from tacobi.view.view_models.view import View

__all__ = ["BaseView", "MaterializedView", "View"]



"""Materialized and normal views that can be used to query cached data."""

from tacobi.view.view_manager import ViewManager
from tacobi.view.view_models import BaseView, MaterializedView, View

__all__ = ["BaseView", "MaterializedView", "View", "ViewManager"]
